/*
  Copyright (c) 2010, miya All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

  * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
  * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
  * Neither the name of the copyright holders nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

// SpursEngine用マンデルブロ集合のリアルタイムアニメーションプログラム
// 2010/12/24 ver.1.00
// 2011/02/08 ver.1.01

#include <iostream>
#include <fstream>
#include <memory>
#include <sys/time.h>
#include <stdlib.h>
#include <pthread.h>
#include <X11/Xlib.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <X11/extensions/XShm.h>
#include <X11/Xutil.h>
#include <spurs/spha/spha.h>
#include <spurs/common/byte_order.h>
#include <spurs/sp3/sp3_cmdif.h>

using namespace std;


// 使用SPE数
#define NUM_OF_SPE 4
// 画像の幅
#define XRES 480
// 画像の高さ
#define YRES 272
#define BPP 4
#define COMMAND_QUIT 0xffffffff
#define LOOP 360
#define WINDOW_NAME "SpursEngine Mandelbrot Set"


// パラメータ用データの共用体（4バイト）
typedef union
{
  int i;
  unsigned int ui;
  float f;
} mcl_cf_param_data_t;


// SPE実行スレッド用構造体
typedef struct
{
  // セッション
  spha_session_t * session;
  // スレッド
  spha_thread_t * thread;
  // SPEのmain()に与えるパラメーター
  uint32_t * param;
  // パラメーターのサイズ
  uint32_t param_size;
} spe_thread_args_t;


// SPEプログラムファイル名
const char * spe_program_name = "spe.elf";

// メッセージポートのデータ
uint32_t data_h2s[32] __attribute__((aligned(128)));
uint32_t data_s2h[NUM_OF_SPE][32] __attribute__((aligned(128)));


// 時間計測用関数。Tick値を返す（マイクロ秒）
static inline long long mcl_get_tick_count(void)
{
  struct timeval timeprof;
  gettimeofday(&timeprof, NULL);
  return ((long long)timeprof.tv_sec * (long long)1000000 + (long long)timeprof.tv_usec);
}


// エラーチェック
// 戻り値: 0:正常 1:エラー
static inline int check_st(spha_status_t st, const char * message)
{
  int error = 0;
  if (st)
  {
    printf("Error: %s\n", message);
    printf("SPHA: %s\n", spha_util_result_code_name(st));
    printf("SP3: %s\n", spha_util_sp3_result_code_name(st));
    error = 1;
  }
  return error;
}


// SPE実行スレッド
void * spe_run_thread(void * arg)
{
  spe_thread_args_t * spearg = (spe_thread_args_t *)arg;
  
  // SPEスレッドを実行
  spha_resume_spe_thread(*(spearg->session), *(spearg->thread), SP3_SPE_THREAD_RESUME_FLAG_BLOCKING, SP3_SPE_THREAD_RESUME_DEFAULT_ENTRY, spearg->param, spearg->param_size, NULL, NULL, NULL);

  // スレッド終了
  pthread_exit(NULL);
}


int main(int argc, char ** argv)
{
  spha_session_t session;
  uint32_t access_attributes = SP3_OBJ_ATTR_HOST_READABLE | SP3_OBJ_ATTR_HOST_WRITABLE | SP3_OBJ_ATTR_HOST_EXECUTABLE | SP3_OBJ_ATTR_SPE_READABLE | SP3_OBJ_ATTR_SPE_WRITABLE | SP3_OBJ_ATTR_SPE_EXECUTABLE | SP3_OBJ_ATTR_SPE_MAPPED_READABLE | SP3_OBJ_ATTR_SPE_MAPPED_WRITABLE;
  float zoom, cx, cy, cz, sx, sy, sz, speed;
  unsigned long long start_tick;
  unsigned int color_shift, cs_on, page;
  cx = 0.50485;
  cy = -0.6099;
  cz = 2.0;
  sx = 0.0;
  sy = 0.0;
  sz = 0.0;
  zoom = 2.0;
  speed = 0.97;
  color_shift = 0;
  cs_on = 0;
  page = 0;

  Display * display;
  int screen;
  Window window;
  GC gc;
  XImage * image;
  XShmSegmentInfo shminfo;
  XSetWindowAttributes att;
  int depth;

  // Xサーバーに接続
  display = XOpenDisplay(NULL);

  // デフォルトスクリーンを取得
  screen = DefaultScreen(display);

  // スクリーンの深度を取得
  depth = DefaultDepth(display, screen);

  // ウィンドウを生成
  window = XCreateSimpleWindow(display, RootWindow(display, screen), 0, 0, XRES, YRES, 1, BlackPixel(display, screen), WhitePixel(display, screen));

  // ウィンドウマネージャをバイパス
  //att.override_redirect=True;
  //XChangeWindowAttributes(display, window, CWOverrideRedirect, &att);

  // ウィンドウ名をセット
  XStoreName(display, window, WINDOW_NAME);

  // 共有メモリー上のXImageを生成
  image = XShmCreateImage(display, DefaultVisual(display, screen), depth, ZPixmap, NULL, &shminfo, XRES, YRES);

  // フルカラーモードでなければ終了
  if (image->bitmap_pad != BPP * 8)
  {
    printf("Please execute in full-color mode.\n");
    exit(0);
  }

  // 共有メモリー・セグメントを生成
  shminfo.shmid = shmget(IPC_PRIVATE, image->bytes_per_line * image->height, IPC_CREAT|0777);

  // 共有メモリー・セグメントをアタッチ
  shminfo.shmaddr = image->data = (char*)shmat(shminfo.shmid, 0, 0);

  // Xサーバーからアクセス可能にする
  shminfo.readOnly = False;

  // 共有メモリーをアタッチすることをXサーバーに通知
  XShmAttach(display, &shminfo);

  // 出力バッファのsync
  XSync(display, False);

  // バッキングストアを有効にする
  att.backing_store = WhenMapped;
  XChangeWindowAttributes(display, window, CWBackingStore, &att);

  // GCを取得
  gc = DefaultGC(display, screen);

  // Copyする際にExposeイベントを送らない
  XSetGraphicsExposures(display, gc, False);

  // 描画色をセット
  XSetForeground(display, gc, BlackPixel(display, screen));

  // Exposeイベントを検出する
  XSelectInput(display, window, ExposureMask);

  // ウィンドウをマップ
  XMapWindow(display, window);

  // 出力バッファのflush
  XFlush(display);

  // セッションの作成
  check_st(spha_create_session(NULL, &session), "create");

  // 通信路を確立
  check_st(spha_connect_session(session), "connect");

  // リソースを予約
  spha_resource_description_t res;
  res.num_spe = NUM_OF_SPE;
  res.num_decm = 0;
  res.num_dech = 0;
  res.num_encm = 0;
  res.num_ench = 0;
  check_st(spha_reserve_resources(session, 100000, 0, &res), "reserve");

  // SPEプログラムファイルをオープン
  ifstream fp(spe_program_name, ios::binary);

  // プログラムサイズを調べる
  fp.seekg(0, ios::end);
  uint32_t spe_elf_size = fp.tellg();
  fp.seekg(0, ios::beg);

  // プログラムサイズを128の倍数に揃える
  spe_elf_size = (spe_elf_size + 127) & ~127;

  // ホスト側バッファを確保（128バイト・アライン）
  char * buffer_elf;
  int err = posix_memalign((void **)&buffer_elf, 128, spe_elf_size);
  if (err)
  {
    abort();
  }

  // SPEプログラムファイルをロード
  fp.read(buffer_elf, spe_elf_size);

  // SpursEngine側ローカルメモリ領域の確保
  // プログラム用
  spha_object_t buffer_obj_elf;
  spha_create_memory(session, access_attributes, 0, 0, spe_elf_size, 0, &buffer_obj_elf);
  // データ用
  uint32_t buffer_size = 0x1800000;
  spha_object_t buffer_obj;
  spha_create_memory(session, access_attributes, 0, 0, buffer_size, 0, &buffer_obj);

  // メモリオブジェクトをマップ
  uint32_t buffer_ea, buffer_ea_elf;
  spha_map_memory(session, buffer_obj_elf, 0, 0, 0, spe_elf_size, 0, &buffer_ea_elf);
  spha_map_memory(session, buffer_obj, 0, 0, 0, buffer_size, 0, &buffer_ea);

  // SpursEngine側ローカルメモリに転送
  uint32_t transfer_size = spe_elf_size;
  spha_data_transfer_to(session, buffer_obj_elf, 0, buffer_elf, &transfer_size);

  // SPEプログラムローダに与える引数をセット
  // エンディアン変換して渡す
  uint32_t loader_arg[4];
  // プログラムデータの実効アドレス
  loader_arg[0] = spurs_convert_ui32(buffer_ea_elf);
  // プログラムデータのサイズ
  loader_arg[1] = spurs_convert_ui32(spe_elf_size);
  loader_arg[2] = 0;
  loader_arg[3] = 0;

  // メッセージポート作成
  // mport: ポート名(ID)
  // seq: シーケンス番号
  // h2sはホストからSPEへの送信用
  // s2hはSPEからホストへの送信用
  spha_object_t mport_h2s, mport_s2h[NUM_OF_SPE];
  uint32_t mport_data_size = 128;
  uint32_t seq_h2s, seq_s2h[NUM_OF_SPE];
  spha_create_message_port(session, access_attributes, mport_data_size, &mport_h2s);
  seq_h2s = 0;
  for (int i = 0; i < NUM_OF_SPE; i++)
  {
    spha_create_message_port(session, access_attributes, mport_data_size, &mport_s2h[i]);
    seq_s2h[i] = 0;
  }

  // SPEのmain()に与えるパラメーター
  uint32_t arg[NUM_OF_SPE][12];

  spha_thread_t spe_th[NUM_OF_SPE];
  for (int i = 0; i < NUM_OF_SPE; i++)
  {
    // SPEスレッドを作成
    spha_create_spe_thread(session, access_attributes, 0, SP3_SPE_LOADER_DEFAULT, loader_arg, sizeof(loader_arg), &spe_th[i]);
  }

  spe_thread_args_t spe_arg[NUM_OF_SPE];
  pthread_t pth[NUM_OF_SPE];
  for (int i = 0; i < NUM_OF_SPE; i++)
  {
    // SPEの通し番号
    arg[i][0] = spurs_convert_ui32((uint32_t)i);
    // 全SPE数
    arg[i][1] = spurs_convert_ui32(NUM_OF_SPE);
    // メッセージポートの名前(ID)
    // (ポート名はエンディアン変換しない)
    arg[i][2] = mport_h2s;
    arg[i][3] = mport_s2h[i];
    // メッセージポートのデータのサイズ
    arg[i][4] = spurs_convert_ui32(mport_data_size);

    // SPEスレッドを実行
    spe_arg[i].session = &session;
    spe_arg[i].thread = &spe_th[i];
    spe_arg[i].param = arg[i];
    spe_arg[i].param_size = sizeof(arg);

    pthread_create(&pth[i], NULL, &spe_run_thread, &spe_arg[i]);
  }

  // 時間計測開始
  start_tick = mcl_get_tick_count();

  zoom = 2.0f;
    
  for (int j = 0; j < LOOP; j++)
  {
    zoom *= speed;
    color_shift++;
    // SPEに渡すパラメータをセット
    mcl_cf_param_data_t var;
    // コマンド
    var.ui = 0;
    data_h2s[0] = spurs_convert_ui32(var.ui);
    // 中心のx座標
    var.f = cx;
    data_h2s[1] = spurs_convert_ui32(var.ui);
    // 中心のy座標
    var.f = cy;
    data_h2s[2] = spurs_convert_ui32(var.ui);
    // 計算範囲
    var.f = zoom;
    data_h2s[3] = spurs_convert_ui32(var.ui);
    // 画面のx方向の解像度
    var.ui = XRES;
    data_h2s[4] = spurs_convert_ui32(var.ui);
    // 画面のy方向の解像度
    var.ui = YRES;
    data_h2s[5] = spurs_convert_ui32(var.ui);
    // 色シフト量
    var.ui = color_shift;
    data_h2s[6] = spurs_convert_ui32(var.ui);
    // フレームバッファのアドレス
    var.ui = buffer_ea + page * XRES * YRES * BPP;
    data_h2s[7] = spurs_convert_ui32(var.ui);

    // SPE演算開始
    spha_signal_message_port(session, mport_h2s, SP3_MPORT_SIGNAL_ASYNC, seq_h2s, &data_h2s, mport_data_size, &seq_h2s);

    // SpursEngineローカルメモリ上の
    // フレームバッファを切り替え(0と1で交互に)
    page = page ^ 1;

    // SpursEngine側ローカルメモリからフレームバッファに転送
    transfer_size = XRES * YRES * BPP;
    spha_data_transfer_from(session, buffer_obj, page * XRES * YRES * BPP, image->data, &transfer_size);

    // Imageを書き込む
    XShmPutImage(display, window, gc, image, 0, 0, 0, 0, XRES, YRES, False);
    XFlush(display);
    XSync(display, False);

    // 演算終了を待つ
    for (int i = 0; i < NUM_OF_SPE; i++)
    {
      spha_wait_message_port(session, mport_s2h[i], SP3_MPORT_WAIT_NEXT_SEQ, seq_s2h[i], SP3_TIMEOUT_INDEFINITE, &seq_s2h[i], &data_s2h[i], &mport_data_size);
    }
  }

  // 経過時間表示
  float calctime = (float)(mcl_get_tick_count() - start_tick) / 1000000.0f;
  float fps = (float)LOOP / calctime;
  printf("time: %f sec\n", calctime);
  printf("fps: %f\n", fps);

  // 終了コマンド送信
  data_h2s[0] = spurs_convert_ui32(COMMAND_QUIT);
  spha_signal_message_port(session, mport_h2s, SP3_MPORT_SIGNAL_ASYNC, seq_h2s, &data_h2s, mport_data_size, &seq_h2s);

  // SPE実行スレッドの終了を待つ
  for (int i = 0; i < NUM_OF_SPE; i++)
  {
    pthread_join(pth[i], NULL);
  }

  // SPEスレッドの破棄
  for (int i = 0; i < NUM_OF_SPE; i++)
  {
    spha_delete_spe_thread(session, spe_th[i]);
  }

  // メモリオブジェクトをアンマップ
  spha_unmap_memory(session, buffer_ea_elf, 0);
  spha_unmap_memory(session, buffer_ea, 0);

  // SpursEngine側ローカルメモリ領域を解放
  spha_delete_object(session, buffer_obj_elf);
  spha_delete_object(session, buffer_obj);

  // セッションの通信路を切断
  spha_close_session(session);

  // セッションの破棄
  spha_delete_session(session);

  // ホスト側バッファを解放
  free(buffer_elf);

  // 共有メモリーをデタッチすることをXサーバーに通知
  XShmDetach(display, &shminfo);

  // XImageを解放
  XDestroyImage(image);

  // 共有メモリー・セグメントをデタッチ
  shmdt(shminfo.shmaddr);
  shmctl(shminfo.shmid, IPC_RMID, 0);

  // ウィンドウを解放
  XDestroyWindow(display, window);

  // Xサーバーから切断
  XCloseDisplay(display);

  return 0;
}
