/*
  Copyright (c) 2010, miya All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

  * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
  * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
  * Neither the name of the copyright holders nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

// SpursEngine用マンデルブロ集合のリアルタイムアニメーションプログラム
// 2010/12/24 ver.1.00
// 2011/02/08 ver.1.01

#include <stdio.h>
#include <spurs/spsa/spsa.h>
#include <spu_mfcio.h>
#include <spu_intrinsics.h>


#define COMMAND_QUIT 0xffffffff
// 最内側ループの回数
#define MAX_COUNT 1024


// パラメータ用データの共用体（4バイト）
typedef union
{
  int i;
  unsigned int ui;
  float f;
} mcl_cf_param_data_t;


// バッファ
unsigned char buffer[128 * 1024] __attribute__((aligned(128)));

// メッセージポートのデータ
uint32_t data_h2s[32] __attribute__((aligned(128)));
uint32_t data_s2h[32] __attribute__((aligned(128)));


// 最内ループのマクロ
#define INNERLOOP() {                                   \
    vb = spu_msub(spu_mul(vf2, va), vb, vy);            \
    va = spu_sub(va2, spu_add(vb2, vx));                \
    va2 = spu_mul(va, va);                              \
    vb2 = spu_mul(vb, vb);                              \
    vtmp = spu_add(va2, vb2);                           \
    flag = spu_cmpgt(vf4, vtmp);                        \
    vcount = spu_sub(vcount, spu_and(vui1, flag));      \
    flagui = spu_extract(spu_gather(flag), 0);          \
    if (flagui == 0) break;}


// マンデルブロ集合を計算
void mandel(void * param_arg)
{
  mcl_cf_param_data_t * param = param_arg;
  vector unsigned int vui1 = spu_splats(1u);
  vector unsigned int vmaxcount = spu_splats((unsigned int)MAX_COUNT);
  vector float vf2 = spu_splats(2.0f);
  vector float vf4 = spu_splats(4.0f);
  unsigned int i, j, k, xres, yres, width, height, page, startline, endline, color_shift, my_num, spes_num;
  // LS上のピクセルデータのポインタ（16バイト単位）
  vector unsigned int *pix;
  void *ptmp;
  // メインメモリの実効アドレス（ピクセルデータの書き込み先）
  unsigned long long addr_ea;
  float xmin, xmax, ymin, ymax, dx, dy;
  // エンディアン変換のためのシャッフルパターン
  vector unsigned char vsp_endian = {3,2,1,0,7,6,5,4,11,10,9,8,15,14,13,12};

  // パラメータ読み込み
  // マンデルブロのタスクのパラメータをセット
  // 画面の解像度
  xres = param[4].ui;
  yres = param[5].ui;
  // 計算範囲
  xmin = param[1].f - param[3].f * (float)xres / (float)yres;
  xmax = param[1].f + param[3].f * (float)xres / (float)yres;
  ymin = param[2].f - param[3].f;
  ymax = param[2].f + param[3].f;
  // 色シフト量
  color_shift = param[6].ui;
  // フレームバッファのアドレス
  addr_ea = param[7].ui;
  // SPEの通し番号
  my_num = param[8].ui;
  // 全SPE数
  spes_num = param[9].ui;

  width = xres;
  height = yres;
  startline = my_num;
  endline = yres;

  dx = (xmax - xmin) / (float)width;
  dy = (ymax - ymin) / (float)height;

  // vxmin = {xmin, xmin + dx, xmin + dx * 2, xmin + dx * 3};
  vector float vxmin = spu_splats(xmin);
  vxmin = spu_insert(xmin + dx, vxmin, 1);
  vxmin = spu_insert(xmin + dx * 2, vxmin, 2);
  vxmin = spu_insert(xmin + dx * 3, vxmin, 3);

  // vymin = {ymin, ymin, ymin, ymin};
  vector float vymin = spu_splats(ymin);

  // vdx = {dx * 4, dx * 4, dx * 4, dx * 4};
  vector float vdx = spu_splats(dx * 4);

  // vdy = {dy, dy, dy, dy};
  vector float vdy = spu_splats(dy);

  // ダブルバッファのページをリセット
  page = 0;

  // 書き込み先アドレスを求める
  addr_ea += (unsigned long long)xres * 4 * startline;

  for (i = startline; i < endline; i += spes_num)
  {
    vector float vy = spu_add(vymin, spu_mul(vdy, spu_convtf(spu_splats(i), 0)));

    // ピクセルバッファのポインタをLS上の汎用バッファの先頭にセット
    pix = (vector unsigned int *)buffer;

    // ダブルバッファのためのオフセットをプラス
    // vector unsigned int型のポインタで1024、つまり16KB
    pix += page * 1024;

    ptmp = pix;

    for (j = 0; j < width / 4; j++)
    {
      vector float vtmp;
      vector unsigned int flag;

      vector float vx = spu_add(vxmin, spu_mul(vdx, spu_convtf(spu_splats(j), 0)));

      vector float va = vx;
      vector float vb = vy;

      // va2 = va * va;
      vector float va2 = spu_mul(va, va);

      // vb2 = vb * vb;
      vector float vb2 = spu_mul(vb, vb);

      vector unsigned int vcount = vmaxcount;

      for (k = 0; k < MAX_COUNT; k+=8)
      {
        unsigned int flagui;
        INNERLOOP();
        INNERLOOP();
        INNERLOOP();
        INNERLOOP();
        INNERLOOP();
        INNERLOOP();
        INNERLOOP();
        INNERLOOP();
      }

      // 適当にカラーリングしてLS上のピクセルバッファにピクセルの値を書き込む
      // *pix = ((vcount + color_shift) << 12) + vcount;
      //*pix = spu_shuffle(spu_add(spu_sl(vcount, 12), vcount), spu_splats(0u), vsp_endian);
      //*pix = spu_shuffle(spu_add(spu_sl(spu_add(vcount, spu_splats(color_shift)), 12), vcount), spu_splats(0u), vsp_endian);
      // 発散しない部分にも色を付ける場合
      *pix = spu_shuffle(spu_sel(spu_add(spu_sl(spu_add(vcount, spu_splats(color_shift)), 12), vcount), spu_convtu(vtmp, 10), flag), spu_splats(0u), vsp_endian);

      // LS上のピクセルバッファのポインタを進める
      pix++;
    }

    // ダブルバッファの前回のDMAが完了するまで待つ
    mfc_write_tag_mask(1 << page);
    mfc_write_tag_update_all();
    mfc_read_tag_status();

    // ローカルメモリのフレームバッファに書き込み
    // 書き込みDMA実行。転送サイズは16バイト単位に補正
    mfc_put(ptmp, addr_ea, (width * 4) & ~15, page, 0, 0);

    // ダブルバッファを切り替え(0と1で交互に)
    page = page ^ 1;

    // 書き込み先の実効アドレスを1ライン分進める
    addr_ea += (unsigned long long)xres * 4 * spes_num;
  }

  // ダブルバッファを切り替え
  page = page ^ 1;

  // 最後のDMAが完了するまで待つ
  mfc_write_tag_mask(1 << page);
  mfc_write_tag_update_all();
  mfc_read_tag_status();
}


int main(vector unsigned int arg0, vector unsigned int arg1, vector unsigned int arg2)
{
  // このSPEの通し番号
  unsigned int my_num = spu_extract(arg0, 0);
  // 全SPE数
  unsigned int spes_num = spu_extract(arg0, 1);
  // ホストからのメッセージポート
  spsa_object_t mport_h2s = spu_extract(arg0, 2);
  // このSPEからホストへのメッセージポート
  spsa_object_t mport_s2h = spu_extract(arg0, 3);
  // メッセージポートのデータサイズ
  uint32_t mport_data_size = spu_extract(arg1, 0);
  // メッセージポートのシーケンス番号
  uint32_t seq_h2s = 0, seq_s2h = 0;

  while (1)
  {
    // メッセージ受信(パラメーター読み込み)
    spsa_wait_message_port(mport_h2s, SP3_MPORT_WAIT_NEXT_SEQ, seq_h2s, SP3_TIMEOUT_INDEFINITE, &seq_h2s, data_h2s, &mport_data_size);

    if (data_h2s[0] == COMMAND_QUIT)
    {
      // 終了コマンドが送られてきたら抜ける
      break;
    }

    data_h2s[8] = my_num;
    data_h2s[9] = spes_num;
    mandel(data_h2s);

    data_s2h[0] = data_h2s[0];

    // メッセージ送信(Hostにタスク完了を通知)
    spsa_signal_message_port(mport_s2h, SP3_MPORT_SIGNAL_ASYNC, seq_s2h, data_s2h, mport_data_size, &seq_s2h);
  }

  return 0;
}
